from unittest.mock import call, patch

from django.urls import reverse
from rest_framework import status

from arkindex.process.models import Worker
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


class TestWorkerConfigurations(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.worker_1 = Worker.objects.get(slug="reco")
        cls.worker_2 = Worker.objects.get(slug="dla")
        cls.worker_config = cls.worker_1.configurations.create(name="config time", configuration={"key": "value"})

    def test_list_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:worker-configurations", kwargs={"pk": str(self.worker_1.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:worker-configurations", kwargs={"pk": str(self.worker_1.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.users.utils.get_max_level", return_value=None)
    def test_list_no_rights(self, max_level_mock):
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:worker-configurations", kwargs={"pk": str(self.worker_1.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {"detail": "You do not have a contributor access to this worker."})
        self.assertEqual(max_level_mock.call_args_list, [
            call(self.user, self.worker_1),
        ])

    def test_list(self):
        config_1 = self.worker_2.configurations.create(name="config_1", configuration={"key": "value"})
        config_2 = self.worker_2.configurations.create(name="config_2", configuration={"dulce": "et decorum est"})
        # Only non-archived configurations should be listed
        self.worker_2.configurations.create(
            name="old_thing",
            configuration={"deprecated": "config"},
            archived=True,
        )

        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:worker-configurations", kwargs={"pk": str(self.worker_2.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertListEqual(response.json()["results"], [{
            "id": str(config_1.id),
            "name": "config_1",
            "configuration": {"key": "value"},
            "archived": False,
        }, {
            "id": str(config_2.id),
            "name": "config_2",
            "configuration": {"dulce": "et decorum est"},
            "archived": False,
        }])

    def test_list_archived(self):
        """
        Toggling the archived filter should only show archived configurations
        """
        self.assertTrue(self.worker_1.configurations.filter(archived=False).exists())
        archived_config = self.worker_1.configurations.create(
            name="old_thing",
            configuration={"deprecated": "config"},
            archived=True,
        )
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:worker-configurations", kwargs={"pk": str(self.worker_1.id)}), {"archived": "true"})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertListEqual(response.json()["results"], [{
            "id": str(archived_config.id),
            "name": "old_thing",
            "configuration": {"deprecated": "config"},
            "archived": True,
        }])

    def test_create_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(reverse("api:worker-configurations", kwargs={"pk": str(self.worker_1.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.post(reverse("api:worker-configurations", kwargs={"pk": str(self.worker_1.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.users.utils.get_max_level", return_value=None)
    def test_create_no_rights(self, max_level_mock):
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.post(reverse("api:worker-configurations", kwargs={"pk": str(self.worker_1.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {"detail": "You do not have an execution access to this worker."})
        self.assertEqual(max_level_mock.call_args_list, [
            call(self.user, self.worker_1),
        ])

    def test_create_empty(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(reverse("api:worker-configurations", kwargs={"pk": str(self.worker_1.id)}))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "name": ["This field is required."],
            "configuration": ["This field is required."]
        })

    def test_create_name_already_exists(self):
        name = "config-name"

        self.worker_2.configurations.create(name=name, configuration={"key": "value"})

        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:worker-configurations", kwargs={"pk": str(self.worker_2.id)}),
                data={"name": name, "configuration": {"key": "value", "cahuete": "bidule"}},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {"name": ["A worker configuration with this name already exists for this worker."]}
        )

    def test_create_configuration_already_exists(self):
        config = {"key": "value"}

        test_config = self.worker_2.configurations.create(name="config-name", configuration=config)

        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:worker-configurations", kwargs={"pk": str(self.worker_2.id)}),
                data={"name": "New configuration", "configuration": config},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {
                "configuration": ["A worker configuration with this configuration already exists for this worker."],
                "id": [str(test_config.id)],
            },
        )

    def test_create_configuration_and_name_already_exist_together(self):
        config = {"key": "value"}
        test_config = self.worker_2.configurations.create(name="a config name", configuration=config)

        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:worker-configurations", kwargs={"pk": str(self.worker_2.id)}),
                data={"name": "a config name", "configuration": config},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {
                "configuration": ["A worker configuration with this configuration already exists for this worker."],
                "id": [str(test_config.id)],
                "name": ["A worker configuration with this name already exists for this worker."]
            }
        )

    def test_create_configuration_and_name_already_exist_separately(self):
        config_1 = {"key": "value"}
        config_2 = {"a": 2, "b": "cinq"}
        self.worker_2.configurations.create(name="a config name", configuration=config_1)
        test_config_2 = self.worker_2.configurations.create(name="some config", configuration=config_2)

        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:worker-configurations", kwargs={"pk": str(self.worker_2.id)}),
                data={"name": "a config name", "configuration": config_2},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {
                "configuration": ["A worker configuration with this configuration already exists for this worker."],
                "id": [str(test_config_2.id)],
                "name": ["A worker configuration with this name already exists for this worker."]
            }
        )

    def test_create(self):
        """
        A user is able to create a worker configuration if they have an admin access on it or its repository
        """
        name = "New configuration"
        config = {"key": "value", "cahuete": "bidule"}

        self.worker_2.configurations.create(name="config_1", configuration={"key": "value"})

        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:worker-configurations", kwargs={"pk": str(self.worker_2.id)}),
                data={"name": name, "configuration": config},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        configuration = self.worker_2.configurations.get(name=name)
        self.assertDictEqual(response.json(), {
            "id": str(configuration.id),
            "name": name,
            "configuration": config,
            "archived": False,
        })

    def test_create_no_config(self):
        name = "New configuration"

        self.worker_2.configurations.create(name="config_1", configuration={"key": "value"})

        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:worker-configurations", kwargs={"pk": str(self.worker_2.id)}),
                data={"name": name},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"configuration": ["This field is required."]})

    def test_create_not_json(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:worker-configurations", kwargs={"pk": str(self.worker_1.id)}),
                data={"name": "test_config", "configuration": []}, format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "configuration": ['Expected a dictionary of items but got type "list".']
        })

    def test_create_null_config(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:worker-configurations", kwargs={"pk": str(self.worker_1.id)}),
                data={"name": "test_config", "configuration": None}, format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "configuration": ["This field may not be null."]
        })

    def test_create_empty_config(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:worker-configurations", kwargs={"pk": str(self.worker_1.id)}),
                data={"name": "test_config", "configuration": {}}, format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "configuration": ["This dictionary may not be empty."]
        })

    def test_retrieve_invalid_id(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.get(
                reverse("api:configuration-retrieve", kwargs={"pk": "12341234-1234-1234-1234-123412341234"})
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_retrieve_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:configuration-retrieve", kwargs={"pk": str(self.worker_config.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_retrieve_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:configuration-retrieve", kwargs={"pk": str(self.worker_config.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.users.utils.get_max_level", return_value=None)
    def test_retrieve_no_rights(self, max_level_mock):
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:configuration-retrieve", kwargs={"pk": str(self.worker_config.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {"detail": "You do not have a contributor access to this worker."})
        self.assertEqual(max_level_mock.call_args_list, [
            call(self.user, self.worker_1),
        ])

    def test_retrieve(self):
        self.worker_config.worker.memberships.get_or_create(user=self.user, level=Role.Guest.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:configuration-retrieve", kwargs={"pk": str(self.worker_config.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "id": str(self.worker_config.id),
            "name": "config time",
            "configuration": {"key": "value"},
            "archived": False,
        })

    def test_update_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.put(
                reverse("api:configuration-retrieve", kwargs={"pk": str(self.worker_config.id)}),
                data={"archived": True}
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_update_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.put(
                reverse("api:configuration-retrieve", kwargs={"pk": str(self.worker_config.id)}),
                data={"archived": True}
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.users.utils.get_max_level", return_value=None)
    def test_update_requires_contributor(self, max_level_mock):
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.put(
                reverse("api:configuration-retrieve", kwargs={"pk": str(self.worker_config.id)}),
                data={"archived": True}
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {
            "detail": "You do not have an execution access to this worker.",
        })
        self.assertEqual(max_level_mock.call_args_list, [
            call(self.user, self.worker_1),
        ])

    def test_update(self):
        """
        Can update a configuration with contributor rights on the worker
        """
        self.client.force_login(self.user)
        self.assertFalse(self.worker_config.archived)

        with self.assertNumQueries(5):
            response = self.client.put(
                reverse("api:configuration-retrieve", kwargs={"pk": str(self.worker_config.id)}),
                data={"archived": True, "name": "new name"}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.worker_config.refresh_from_db()
        self.assertEqual(self.worker_config.name, "new name")
        self.assertDictEqual(self.worker_config.configuration, {"key": "value"})
        self.assertTrue(self.worker_config.archived)

    def test_update_ignored_fields(self):
        """
        Fields that should not be editable are ignored when sent in update requests
        """
        # Get as much rights as possible so that this test does not fail if something goes wrong with the endpoint's
        # permissions; we do not care about rights in this test.
        self.client.force_login(self.superuser)
        self.worker_1.memberships.update_or_create(user=self.user, defaults={"level": Role.Admin.value})

        self.assertFalse(self.worker_config.archived)

        with self.assertNumQueries(5):
            response = self.client.put(
                reverse("api:configuration-retrieve", kwargs={"pk": str(self.worker_config.id)}),
                data={
                    "name": "new name",
                    "configuration": {"stop": "hammertime"},
                    "archived": True,
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.worker_config.refresh_from_db()
        self.assertEqual(self.worker_config.name, "new name")
        self.assertTrue(self.worker_config.archived)
        # The configuration was not updated
        self.assertDictEqual(self.worker_config.configuration, {"key": "value"})

    def test_update_archived_configuration_name(self):
        """
        Archived configurations cannot be renamed with an update
        """
        self.client.force_login(self.user)
        self.worker_config.archived = True
        self.worker_config.save()
        self.assertTrue(self.worker_config.archived)

        with self.assertNumQueries(4):
            response = self.client.put(
                reverse("api:configuration-retrieve", kwargs={"pk": str(self.worker_config.id)}),
                data={
                    "name": "new name",
                    "archived": True,
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "name": ["Archived configurations cannot be renamed."],
        })

        self.worker_config.refresh_from_db()
        self.assertEqual(self.worker_config.name, "config time")
        self.assertTrue(self.worker_config.archived)

    def test_update_archived_configuration_name_unarchive(self):
        """
        Archived configurations can be renamed if they are also unarchived
        """
        self.client.force_login(self.user)
        self.worker_config.archived = True
        self.worker_config.save()
        self.assertTrue(self.worker_config.archived)

        with self.assertNumQueries(5):
            response = self.client.put(
                reverse("api:configuration-retrieve", kwargs={"pk": str(self.worker_config.id)}),
                data={
                    "name": "new name",
                    "archived": False,
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.worker_config.refresh_from_db()
        self.assertEqual(self.worker_config.name, "new name")
        self.assertFalse(self.worker_config.archived)

    def test_update_name_exists(self):
        """
        Updating the name of a configuration to the name of another existing configuration on the same worker
        should fail as this would break the unique constraint
        """
        self.worker_1.configurations.create(name="new name", configuration={"key": "value", "key2": "value2"})
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.put(
                reverse("api:configuration-retrieve", kwargs={"pk": str(self.worker_config.id)}),
                data={
                    "name": "new name",
                    "archived": False,
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "name": ["A worker configuration with this name already exists for this worker."],
        })

    def test_partial_update_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.patch(
                reverse("api:configuration-retrieve", kwargs={"pk": str(self.worker_config.id)}),
                data={"archived": True}
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_partial_update_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.patch(
                reverse("api:configuration-retrieve", kwargs={"pk": str(self.worker_config.id)}),
                data={"archived": True}
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.users.utils.get_max_level", return_value=None)
    def test_partial_update_requires_contributor(self, max_level_mock):
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            response = self.client.patch(
                reverse("api:configuration-retrieve", kwargs={"pk": str(self.worker_config.id)}),
                data={"archived": True}
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(response.json(), {
            "detail": "You do not have an execution access to this worker.",
        })
        self.assertEqual(max_level_mock.call_args_list, [
            call(self.user, self.worker_1),
        ])

    def test_partial_update(self):
        """
        Can partial update a configuration with contributor rights on the worker
        """
        self.worker_1.memberships.all().delete()
        self.worker_1.memberships.create(user=self.user, level=Role.Contributor.value)
        self.client.force_login(self.user)

        self.assertFalse(self.worker_config.archived)

        with self.assertNumQueries(4):
            response = self.client.patch(
                reverse("api:configuration-retrieve", kwargs={"pk": str(self.worker_config.id)}),
                data={"archived": True}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.worker_config.refresh_from_db()
        self.assertEqual(self.worker_config.name, "config time")
        self.assertDictEqual(self.worker_config.configuration, {"key": "value"})
        self.assertTrue(self.worker_config.archived)

    def test_partial_update_ignored_fields(self):
        """
        Fields that should not be editable are ignored when sent in partial update requests
        """
        # Get as much rights as possible so that this test does not fail if something goes wrong with the endpoint's
        # permissions; we do not care about rights in this test.
        self.client.force_login(self.superuser)
        self.worker_1.memberships.update_or_create(user=self.user, defaults={"level": Role.Admin.value})

        self.assertFalse(self.worker_config.archived)

        with self.assertNumQueries(5):
            response = self.client.patch(
                reverse("api:configuration-retrieve", kwargs={"pk": str(self.worker_config.id)}),
                data={
                    "name": "new name",
                    "configuration": {"stop": "hammertime"},
                    "archived": True,
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.worker_config.refresh_from_db()
        self.assertEqual(self.worker_config.name, "new name")
        self.assertTrue(self.worker_config.archived)
        # The configuration was not updated
        self.assertDictEqual(self.worker_config.configuration, {"key": "value"})

    def test_partial_update_archived_configuration_name(self):
        """
        Archived configurations cannot be renamed with a partial update
        """
        self.client.force_login(self.user)
        self.worker_config.archived = True
        self.worker_config.save()
        self.assertTrue(self.worker_config.archived)

        with self.assertNumQueries(4):
            response = self.client.patch(
                reverse("api:configuration-retrieve", kwargs={"pk": str(self.worker_config.id)}),
                data={
                    "name": "new name",
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "name": ["Archived configurations cannot be renamed."],
        })

        self.worker_config.refresh_from_db()
        self.assertEqual(self.worker_config.name, "config time")
        self.assertTrue(self.worker_config.archived)

    def test_partial_update_name_exists(self):
        """
        Updating the name of a configuration to the name of another existing configuration on the same worker
        should fail as this would break the unique constraint
        """
        self.worker_1.configurations.create(name="new name", configuration={"key": "value", "key2": "value2"})
        self.client.force_login(self.user)

        with self.assertNumQueries(4):
            response = self.client.patch(
                reverse("api:configuration-retrieve", kwargs={"pk": str(self.worker_config.id)}),
                data={
                    "name": "new name",
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            "name": ["A worker configuration with this name already exists for this worker."],
        })
