from typing import Optional

from django.db.models import IntegerField, Value

from arkindex.users.models import Role, User


def has_access(user: User, instance, level: int, skip_public: bool = False) -> bool:
    """
    Check if the user has access to a generic instance with a minimum level
    If skip_public parameter is set to true, exclude rights on public instances
    """
    if user.is_anonymous:
        return level <= Role.Guest.value and not skip_public and getattr(instance, "public", False)
    return True


def filter_rights(user: User, model, level: int):
    """
    Return a generic queryset of objects with access rights for this user.
    Level filtering parameter should be an integer between 1 and 100.
    """
    if user.is_anonymous:
        if hasattr(model, "public"):
            return model.objects.filter(public=True).annotate(max_level=Value(Role.Guest.value, IntegerField()))
        return model.objects.none()

    return model.objects.annotate(max_level=Value(Role.Admin.value, IntegerField()))


def get_max_level(user: User, instance) -> Optional[int]:
    """
    Returns the maximum access level on a given model instance
    """
    if user.is_anonymous:
        if getattr(instance, "public", False):
            return Role.Guest.value
        return None

    return Role.Admin.value
